/********************************************************************
 * C Users Group (U.K) C Source Code Library File CUGLIB.018        *
 * Inquiries to: M. Houston, 36 Whetstone Clo. Farquhar Rd.         *
 * Edgbaston, Birmingham B15 2QN ENGLAND			    *
 ********************************************************************
 * File name: curses.c
 * Program name: curses screen control library
 * Source of file: West Midlands OPUS BBS
 * Purpose: An MS-DOS implementation of the UNIX curses screen
 *          management library.
 * Changes: <who what when & why major changes have been made>      
 ********************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <curses.h>  /* routines in this file.           */
#include <box.h>     /* of IBM box-drawing characters    */
#include <stdarg.h>  /* va_list and va_start (ANSI)      */

/*-------------------------------------------------------*
 * CURSES.C:  Copyright (c) 1987, Allen I. Holub.        *
 *                  All rights reserved.                 *
 *-------------------------------------------------------*
 *
 * This file is a DOS implementation of some of the Unix
 * CURSES functions. It is Unix compatible but is a proper
 * subset, not a full implementation, of curses. It works
 * on the IBM-PC. In all of these y is the row number and
 * x is the column number. The upper left corner is (0,0):
 *
 * I/O functions ------------------------------------------
 *
 * waddch( win, ch )      Works like putc()
 * waddstr( win, str)     Works like puts()
 * wprintw(win,fmt,arg...)  Like printf but writes to the
 *                          indicated window.
 * wrefresh(win)          See below.
 * box(win, vert, horiz)  Draws a box around the window.
 *
 * Cursor movement and screen control ---------------------
 *
 * werase(win)       erase the entire window
 * wclrtoeol(win)    erase from cursor position to the end
 *                   of line
 *
 * wmove(win,y,x)  Move the cursor to postion (y,x) rela-
 *                 tive to the orgin of the indicated window.
 *
 * getxy(win,y,x)    MACRO: puts the current cursor postion
 *                   into y and x. Note that this is a macro,
 *                   don't put an & in front of y and x in
 *                   the invocation.
 *
 * wgetch(win)       works like getchar but echos to the
 *                   indicated window.  If crmode is
 *                   inactive, it is activated for the
 *                   duration of this subroutine.
 *
 * scroll(win)          Scrolls the window up one line.
 * wscroll( win, amt )  NOT A CURSES FUNCTION. Scrolls
 *                      window by indicated amount. A
 *                      negaitve amt scrolls down.
 *
 * Initialization stuff -------------------------------------
 *
 * initscr()              Initialize
 * endwin()               Clean up
 * scrollok(win, flag);   enable/disable scrolling for window.
 * newwin(lines,cols,begin_y,begin_x)     Create a new window.
 *
 * Terminal control -----------------------------------------
 *        Because of the perversities of DOS, these work in a
 *        slightly nonstandard way. In particular echo, noecho,
 *        nl, and nonl only work if crmode() is active.
 *        Moreover, they are ignored for the non-w fundtions.
 *        For portability reasons, it's best to always set
 *        crmode() at the top of your program.
 *
 * crmode();              Turn off input buffering.
 * nocrmode();            Turn it back on again.  (default)
 * noecho();              Turn off automatic echo.
 * echo();                Turn it back on again.  (default)
 * nl();                  Turn on CR-NL mapping   (default)
 * nonl();                turn it off again.
 *
 * Functions that affect the whole screen. ------------------
 *
 * move(y,x)              move cursor to abs. position (y,x)
 * addch(c)               Write a character.
 * clear()                Clear the screen
 * printw()               works like printf
 * getch()                get a character from the keyboard.
 * refresh()              See below.
 *
 *-----------------------------------------------------------
 * The real curses keeps an two internal representations of
 * the screen, when you change something it just modifies
 * one of these reprsentations. You must issue a refresh()
 * or wrefresh() call to actually modify the screen. My
 * version of curses refreshes the screen immediately after
 * every write. refresh() and wrefresh() macros have been
 * provided for UNIX compatability, however. These macros
 * don't do anything, but you should scatter them liberally
 * about your code if you want it to be portable.
 *
 * I've corrected one bug in the real curses that might
 * cause problems when you port your code. The real curses
 * (at least the one at Berkeley), doesn't scroll properly
 * in that it leaves junk on the bottom line of the window
 * after a scroll. I've corrected the problem here but,
 * again, if you want real portability you should do a
 * wclrtoeol(win) after every scroll. Unfortunately, there's
 * no way to determine that the screen has scrolled without
 * actually keeping track of the characters that are written
 * to the screen. Ugh.
 *
 * Other differences: curses doesn't know about characters
 * that it hasn't actually put on the screen with an addch().
 * So, if echo is enabled, curses won't erase the echoed
 * characters when it scrolls the screen and so forth. The
 * curses presented here doesn't exibit this behaviour, but
 * if you want compatablility with Unix, turn off echo (with
 * a noecho() call) and then echo all characters yourself.
 *
 * The purpose of the four #defines immediately below is to
 * make it easy to modify this package. They all map function
 * calls to video-bios subroutines [vb_xxx()]. If you want to
 * use your own functions (that do direct video access or send
 * out escape sequences to a normal terminal, for example)
 * just change the #defines and recompile. The functions must
 * behave as follows:
 *
 * cmove(y,x)   Move the cursor to position (y,x) (y is row)
 *              where (0,0) is the upper-left corner of the
 *              screen.
 * curpos(y,x)  Put the (y,x) cursor position into the
 *              integers pointed to by y and x.
 * replace(c)   Print c at the current cursor position without
 *              moving the cursor. This routine must be able
 *              to put a character into the 80th column
 *              without scrolling the screen or wrapping
 *              around.
 *
 * doscroll(l,r,t,b,a)  Scroll a region of the screen with the
 *                      top left corner at (t,l) and the bottom
 *              right corner at (b,r). "a" is the number of
 *              lines to scroll (up if a is positive, down if
 *              it's negative).
 *
 * inchar()     Returns the character at the current curser
 *              position.
 */

#define cmove(y,x)           vb_ctoyx(y,x);
#define curpos(y,x)          vb_getyx(y,x);
#define replace(c)           vb_replace(c);
#define doscroll(l,r,t,b,a)  vb_scroll(l,r,t,b,a);
#define inchar()             vb_inchar(NULL)

#define min(a,b)             ((a) < (b) ? (a) : (b))

/*--------------------------------------------------------*/

static int Echo   = 1;  /* Echo enabled                   */
static int Crmode = 0;  /* If 1, use buffered input       */
static int Nl     = 1;  /* If 1, map \r to \n on input    */
                        /* and map both to \n\r on output */

/*--------------------------------------------------------*/

noecho  (     ) {  Echo = 0;                    }
echo    (     ) {  Echo = 1;                    }
nl      (     ) {  Nl   = 1;                    }
nonl    (     ) {  Nl   = 0;                    }
getpos  (yp,xp) {  return curpos(yp, xp);       }
move    (y, x ) {  cmove(y, x);                 }
inch    (     ) {  return inchar();             }

 /*--------------------------------------------------------*/

crmode()
{
    FILE        *console;

    setvbuf(stdin, NULL, _IONBF, 0);  /* Turn off buffering */
    Crmode = 1;
}

nocrmode()
{
    freopen( "/dev/con", "r", stdin );
    Crmode = 0;
}

/*--------------------------------------------------------*/

WINDOW  *newwin( lines, cols, begin_y, begin_x )

int     cols;     /* Horizontal size (including border)   */
int     lines;    /* Vergtical size  (including bofder)   */
int     begin_y;  /* X coordinate of upper-left corner    */
int     begin_x;  /* Y coordinate of upper-left corner    */
{
        WINDOW  *win, *malloc();

        if( !(win = malloc( sizeof(WINDOW) )) )
                perror("Out of memory\n");

        win->x_org      = begin_x;
        win->y_org      = begin_y;
        win->x_size     = cols   ;
        win->y_size     = lines  ;
        win->row        = 0;
        win->col        = 0;
        win->scroll_ok  = 0;

        werase(win);
        return win;
}

/*--------------------------------------------------------*/

int     waddch( win, c )
WINDOW  *win;
int     c;
{
    /*  Print a character: The following are handled
     *  specially:
     *
     *  \n  Clear the line from the current cursor position
     *      to the right edge of the window. Then:
     *      if nl() is active:
     *          go to the left edge of the next line
     *      else
     *          go to the current column on the next line
     *      In addition, if scrolling is enabled, the window
     *      scrolls if you're on the bottom line.
     *  \t  is expanded into an 8-space field. If the tab
     *      goes past the right edge of the window, the
     *      cursor wraps to the next line.
     *  \r  gets you to the beginning of the current line.
     *
     *  \b  backs up one space but may not back up past
     *      the left edge of the window. Nondestructive. The
     *      curses documentation doesn't say that \b is
     *      handled explicitly but it does indeed work.
     *
     *  The following is not supported by Unix. Don't use
     *  explicit escape sequences if portability is a
     *  consideration:
     *
     *  ESC This is not standard curses but is useful. All
     *      characters between an ASCII ESC and an alphabetic
     *      character are sent to the output but are otherwise
     *      igonred. This let's you send escape sequences
     *      directly to the terminal if you like. I'm
     *      assuming here that you won't change windows in the
     *      middle of an escape sequence.
     *
     * Return ERR if the character would have caused the
     * window to scroll illegally.
     */

    static  int     saw_esc = 0;
    static  WINDOW  *oldwin = NULL;
    int             rval    = OK;

    if( oldwin != win )
    {
        cmove(win->y_org + win->row, win->x_org + win->col);
        oldwin = win;
    }

    if( saw_esc )
    {
            if( isalpha( c ) )
                    saw_esc = 0;

            putchar(c);
    }
    else
    {
            switch( c )
            {
            case '\033':
                    saw_esc = 1;
                    putchar('\033');
                    break;

            case '\b':
                    if( win->col > 0 )
                    {
                            putchar('\b');
                            --( win->col );
                    }
                    break;

            case '\t':
                    do
                    {
                            waddch( win, ' ' );
                    }
                    while( win->col % 8 );
                    break;

            case '\r':
                    win->col = 0;
                    cmove(win->y_org + win->row, win->x_org);
                    break;

            default :
                    putchar(c);
                    if( ++(win->col) < win->x_size )
                            break;

                    /* fall through to newline */

            case '\n':
                    wclrtoeol( win );

                    if( Nl )
                        win->col = 0;

                    if( ++(win->row) >= win->y_size )
                    {
                        rval = wscroll( win, 1 );
                        --( win->row );
                    }

                    cmove( win->y_org + win->row,
                           win->x_org + win->col);
                    break;
            }
    }

    return rval;
}

/*--------------------------------------------------------*/

waddstr( win, str )
WINDOW  *win;
char    *str;
{
        while( *str )
                waddch( win, *str );
}

/*--------------------------------------------------------*/

static int Errcode = OK;

static wputc(c, win) WINDOW *win;
{
        Errcode |= waddch(win,c);
}

wprintw(win, fmt)
WINDOW  *win;
char    *fmt;
{
    /* The doprnt() function used here is explained in depth
     * in Allen Holub, The C Companion (Prentice-Hall: 1987),
     * pp. 213-237. If you don't have this book, an alternate
     * approach is explained in the text.
     */

    va_list args;
    va_start( args, fmt );

    Errcode = OK;
    doprnt( wputc, win, fmt, args );
    return Errcode;
}

/*--------------------------------------------------------*/

box( win, vert, horiz )
WINDOW  *win;
{
    /* Draws a box in the outermost characters of the window
     * using vert for the vertical characters and horiz for
     * the horizontal ones. I've extended this function to
     * support the IBM box-drawing characters. That is,
     * if IBM box-drawing characters are specified for vert
     * and horiz, box() will use the correct box-drawing
     * characters in the corners. These are defined in
     * box.h as:
     *
     *      HORIZ   (0xc4)  single horizontal line
     *      D_HORIZ (0xcd)  double horizontal line.
     *      VERT    (0xb3)  single vertical line
     *      D_VERT  (0xba)  double vertical line.
     */

    int     i, nrows;
    int     ul, ur, ll, lr;

    if( !( (horiz == HORIZ || horiz == D_HORIZ) &&
           (vert  == VERT  || vert  == D_VERT ) ))
    {
            ul = ur = ll = lr = vert ;
    }
    else
    {
        if( vert == VERT )
        {
            if(horiz==HORIZ)
                ul=UL,    ur=UR,    ll=LL,    lr=LR;
            else
                ul=HD_UL, ur=HD_UR, ll=HD_LL, lr=HD_LR;
        }
        else
        {
            if(horiz==HORIZ)
                ul=VD_UL, ur=VD_UR, ll=VD_LL, lr=VD_LR;
            else
                ul=D_UL,  ur=D_UR,  ll=D_LL,  lr=D_LR;
        }
    }

    cmove( win->y_org, win->x_org );

    putchar( ul );                          /* Top line */

    for( i = win->x_size-2; --i >= 0 ; )
            putchar( horiz );

    putchar( ur );
    nrows = win->y_size - 2 ;              /* Two sides */
    i     = win->y_org  + 1 ;

    while( --nrows >= 0 )
    {
            cmove( i,   win->x_org );
            putchar( vert );

            cmove( i++, win->x_org + (win->x_size - 1) );
            putchar( vert );
    }

    cmove(i, win->x_org);               /* Bottom line */
    putchar( ll );  

    for( i = win->x_size-2; --i >= 0 ; )
            putchar( horiz );

    putchar( lr );
}

/*--------------------------------------------------------*/

werase( win )
WINDOW  *win;
{
    vb_scroll( win->x_org, win->x_org + (win->x_size - 1), 
               win->y_org, win->y_org + (win->y_size - 1),
               win->y_size );

    cmove( win->y_org, win->x_org );
    win->row = 0;
    win->col = 0;
}

/*----------------------------------------------------------
 * Scroll the window if scrolling is enabled. Return 1 if we
 * scrolled. (I'm not sure if the Unix function returns 1
 * on a scroll but it's convenient to do it here. Don't
 * assume anything about the return value if you're porting
 * to Unix. Wscroll() is not a curses function. It lets you
 * specify a scroll amount and direction (scroll down by -amt
 * if amt is negative); scroll() is a macro that evaluates to
 * a wscroll call with an amt of 1. Note that the Unix curses
 * gets very confused when you scroll explicitly (using
 * scroll()). In particular, it doesn't clear the bottom line
 * after a scroll but it thinks that it has. Therefore, when
 * you try to clear the bottom line, it thinks that there's
 * nothing there to clear and ignores your wclrtoeol()
 * commands. Same thing happens when you try to print spaces
 * to the bottom line; it thinks that spaces are already there
 * and does nothing. You have to fill the bottom line with
 * non-space characters of some sort, and then erase it.
 */

wscroll(win, amt)
WINDOW  *win;
{
    if( win->scroll_ok )
        doscroll( win->x_org, win->x_org + (win->x_size-1), 
                  win->y_org, win->y_org + (win->y_size-1),
                  amt);

    return win->scroll_ok ;
}

/*--------------------------------------------------------*/

wmove( win, y, x )
WINDOW  *win;
{
    /* Seek into the window. It's not permitted to seek
     * outside of the window area.
     */

    cmove( win->y_org + (win->row = min(y,win->y_size-1)) ,
           win->x_org + (win->col = min(x,win->x_size-1)) );
}

/*--------------------------------------------------------*/

static  getcon( win )
WINDOW  *win;
{
    /* Get a character from DOS without echoing. We need
     * to do this in order to support (echo/noecho). We'll
     * also do noncrmode input buffering here. Maximum input
     * line length is 132 columns.
     *
     * In nocrmode(), DOS is used to get a line and all the
     * normal command-line editing functions are available.
     * Note that since there's no way to turn off echo in
     * this case, characters will be echoed to the screen
     * regardles of the status of echo().
     * In order to retain control of the window, input
     * fetched for wgetch() is always done in crmode, even
     * if Crmode isn't set.
     * If nl() mode is enabled, carriage return (Enter, ^M)
     * and linefeed (^J) are both mapped to '\n', otherwise
     * they are not mapped.
     */

    static unsigned char  buf[ 133 ] = { 133, 0 };
    static unsigned char  *p         = buf;
    static          int   numchars   = 0;
    register        int   c;

    if( Crmode || win )
    {
            if( (c = bdos(8,0,0) & 0xff) == ('Z'-'@') )
                    return EOF ;

            if( c == '\r' &&  Nl)
                    c = '\n' ;

            if( Echo )
            {
                if( win )  waddch ( win, c );
                else       addch  (      c );
            }

            return c;
    }
    else if( numchars )
    {
            --numchars  ;
            return *p++ ;
    }
    else
    {
            bdos(10, buf, 0);
            numchars =  buf[1];
            p        = &buf[2];
    }
}

wgetch (win)  WINDOW *win; { return getcon( win  ); }
getch  (   )               { return getcon( NULL ); }

/*--------------------------------------------------------*/

clear()
{
        doscroll( 0, 79, 0, 24, 25 );
        move( 0, 0 );
}

/*--------------------------------------------------------*/

wclrtoeol( win )
WINDOW  *win;
{
    /* Clear from cursor to end of line, the cursor isn't
     * moved. The main reson that this is included here is
     * because you have to call it after printing every
     * newline in order to compensate for a bug in the real
     * curses. This bug has been corrected in the curses
     * presented here, however, so you don't have to use
     * this routine if you're not interested in portability.
     * Note that you must use a replace function on the
     * rightmost character to prevent scrolling.
     */

    register int    i;

    for( i = win->x_size - win->col - 1; --i >= 0 ;)
            putchar(' ');

    replace(' ');
    cmove( win->y_org + win->row, win->x_org + win->col )
}

/*--------------------------------------------------------*/
#ifdef MAIN

WINDOW  *boxwin( lines, cols, y_start, x_start )
{
    /* This routine works just like the newwin() except that
     * the window has a box around it that won't be destoyed
     * by writes to the window. It accomplishes this feat by
     * creating two windows, one inside the other, with a box
     * drawn around the outer one.
     */

    WINDOW *outer, *inner;

    outer = newwin(lines,   cols,   y_start,   x_start);

#ifdef MSDOS
    box( outer, VERT, HORIZ );
#else
    box( outer,  '|',  '-'  );
#endif

    wrefresh ( outer );
    return newwin( lines-4, cols-4, y_start+2, x_start+2 ); 
}

pattern()
{
    clear();
    printf("01234567890123456789012345678901234567890 0\n");
    printf("01234567890123456789012345678901234567890 1\n");
    printf("01234567890123456789012345678901234567890 2\n");
    printf("01234567890123456789012345678901234567890 3\n");
    printf("01234567890123456789012345678901234567890 4\n");
    printf("01234567890123456789012345678901234567890 5\n");
    printf("01234567890123456789012345678901234567890 6\n");
    printf("01234567890123456789012345678901234567890 7\n");
    printf("01234567890123456789012345678901234567890 8\n");
    printf("01234567890123456789012345678901234567890 9\n");
    printf("01234567890123456789012345678901234567890 10\n");
    printf("01234567890123456789012345678901234567890 11\n");
    printf("01234567890123456789012345678901234567890 12\n");
    printf("01234567890123456789012345678901234567890 13\n");
    printf("01234567890123456789012345678901234567890 14\n");
    printf("01234567890123456789012345678901234567890 15\n");
    printf("01234567890123456789012345678901234567890 16\n");
    printf("01234567890123456789012345678901234567890 17\n");
    printf("01234567890123456789012345678901234567890 18\n");
    printf("01234567890123456789012345678901234567890 19\n");
    printf("01234567890123456789012345678901234567890 20\n");
    printf("01234567890123456789012345678901234567890 21\n");
    printf("01234567890123456789012345678901234567890 22\n");
    printf("          1         2         3         4 \n");
}

main()
{
        /* All coordinates are (y,x) */

        WINDOW  *win1, *win2;
        char    str[128];
        int     c;

        initscr(); /* Initialize curses                  */
        noecho();  /* Echo off (it screws up the screen) */
        crmode();  /* Put terminal into CBREAK mode      */

	clear();
        pattern();

        win1 = boxwin(10, 38,  0, 0 );
        win2 = boxwin(10, 25, 11, 11);

        scrollok( win1, TRUE );
        wprintw(win1, "This is window one, doo wha\n" );
        wrefresh( win1 );
        wprintw(win2, "This is window 2.\nPress a key\n" );
        wrefresh( win2 );

        c = wgetch(win2);
        wmove(win1, 5, 0 );

        wprintw   (win1, "Got %c, 0x%x\n", c, c );
        wrefresh(win1);

        while( (c = wgetch(win1) & 0x7f) != 'q' )
        {
                if( c == 'X'-'@' )
                    wclrtoeol( win1 );
                else
                    waddch( win1, c );

                wrefresh( win1 );
        }

        move(23,0);
        refresh();
        clear();
        endwin();
}
#endif
